#import "CRUDLayer.h"
#import "CRUDObjectProtocol.h"
#import "EOEntity+CRUDLayerExtensions.h"

@implementation CRUDLayer
- initWithModelName:(NSString*)modelName
{
	NSString *path;
	
	[super init];

	//
	// EOModel
	//
	path = [EOModel findPathForModelNamed:modelName];
	if (!path) {
		NSLog(@"ERROR: CRUDLayer: failed on findPathForModelNamed:%@", modelName);
		return nil;
	}
	model = [[EOModel alloc] initWithContentsOfFile:path];
	if (!model) {
		NSLog(@"ERROR: CRUDLayer: failed to open model '%@'", path);
		return nil;
	}
	NSLog(@"CRUDLayer: successfully opened eomodel '%@'", [model path]);

#ifdef CRUD_DEBUG
{
	EOAdaptor *adaptor;
	EOAdaptorContext *adaptorContext;
	EOAdaptorChannel *adaptorChannel;

	//
	NSLog(@"DEBUG: INFO: Test adaptor alloc...");
	adaptor = [EOAdaptor adaptorWithModel:model];
	if (!adaptor) {
		NSLog(@"DEBUG: FAILURE: adaptorWithModel returned nil!");
		return nil;
	}
	NSLog(@"DEBUG: SUCCESS: adaptor = '%@'", [adaptor name]);
	[adaptor setDelegate:self];

	//
	NSLog(@"DEBUG: INFO: Test connectionDictionary...");
	NSLog(@"DEBUG: INFO: connectionDictionary = %@", [adaptor connectionDictionary]);
	if ([adaptor hasValidConnectionDictionary] == NO) {
		NSLog(@"DEBUG: FAILURE: hasValidConnectionDictionary == NO!!!");
		return nil;
	}
	NSLog(@"DEBUG: SUCCESS: connectionDictionary is valid...");

	//
	NSLog(@"DEBUG: INFO: Test create a channel...");
	adaptorContext = [adaptor createAdaptorContext];
	adaptorChannel = [adaptorContext createAdaptorChannel];
	if (!adaptorChannel) {
		NSLog(@"DEBUG: FAILURE: adaptorChannel == nil!!!");
		return nil;
	}
	NSLog(@"DEBUG: SUCCESS: adaptorChannel = %@", adaptorChannel);

	//
	NSLog(@"DEBUG: INFO: Test a database connection...");
	if ([adaptorChannel openChannel] == NO) {
		NSLog(@"DEBUG: FAILURE: openChannel == NO!!!");
		return nil;
	}
	NSLog(@"DEBUG: SUCCESS: channel is now open...");

	//
	NSLog(@"DEBUG: INFO: Test hasOpenChannels...");
	if ([adaptor hasOpenChannels] == NO) {
		NSLog(@"DEBUG: FAILURE: hasOpenChannels == NO!!!");
		return nil;
	}
	NSLog(@"DEBUG: SUCCESS: channels are open...");

	//
	NSLog(@"DEBUG: INFO: Test closeChannel...");
	[adaptorChannel closeChannel];
	NSLog(@"DEBUG: INFO: closeChannel returns void...");

	NSLog(@"DEBUG: SUCCESS? completed debugging...");
}
#endif

	//
	// EOModel Data
	//
	modelData = [[CRUDEOModelData alloc] initWithModelName:modelName];
	if (!modelData) {
		NSLog(@"ERROR: CRUDLayer: failed to alloc/init CRUDEOModelData.");
		return nil;
	}	

	//
	// dataSources
	//
	dataSources = [[NSMutableDictionary alloc] init];
	if (!dataSources) {
		NSLog(@"ERROR: CRUDLayer: failed alloc/init dataSources!");
		return nil;
	}
	
	return self;
}

- (void)dealloc {
	[dataSources release];
	[model release];
	[modelData release];
	[super dealloc];
}

- (EOModel*) model { return model; }
- (CRUDEOModelData*)modelData { return modelData; }
@end

@implementation CRUDLayer (DataSources)
- (EODatabaseDataSource*)dataSourceForEntityNamed:(NSString*)entityName
{
	EODatabaseDataSource *dataSource;
	
	if (!entityName) {
		NSLog(@"dataSourceForEntityNamed: entity = nil!");
		return nil;
	}
	
	dataSource = [dataSources objectForKey:entityName];
	if (dataSource)
		return dataSource;

	NSLog(@"dataSourceForEntityNamed: allocating dataSource for '%@'.", entityName);

	dataSource = [[EODatabaseDataSource alloc]
				initWithModelName:[model name]
				entityName:entityName];
	if (!dataSource) {
		NSLog(@"ERROR: dataSourceForEntityNamed: failed on '%@'!", entityName);
		return nil;
	}
	[dataSources setObject:dataSource forKey:entityName];

#ifdef CRUD_DEBUG
{
	EOAdaptorChannel *adaptorChannel;
	EOAdaptor *adaptor;

	//
	NSLog(@"DEBUG: INFO: create adaptorChannel from dataSource...");
	adaptorChannel = [[dataSource databaseChannel] adaptorChannel];
	if (!adaptorChannel) {
		NSLog(@"DEBUG: FAILURE: adaptorChannel == nil!!!");
		return nil;
	}
	NSLog(@"DEBUG: SUCCESS: adaptorChannel = %@", adaptorChannel);

	//
	NSLog(@"DEBUG: INFO: become the adaptorChannel delegate...");
	[adaptorChannel setDelegate:self];
	NSLog(@"DEBUG: INFO: setDelegate returns void...");


	//
	NSLog(@"DEBUG: INFO: create adaptor from adaptorChannel...");
	adaptor = [[adaptorChannel adaptorContext] adaptor];
	if (!adaptor) {
		NSLog(@"DEBUG: FAILURE: adaptor == nil!!!");
		return nil;
	}
	NSLog(@"DEBUG: SUCCESS: adaptor = '%@'", [adaptor name]);

	//
	NSLog(@"DEBUG: INFO: become the adaptor delegate...");
	[adaptor setDelegate:self];
	NSLog(@"DEBUG: INFO: setDelegate returns void...");

	NSLog(@"DEBUG: SUCCESS? completed debugging...");
}
#endif

	return dataSource;
}
@end

@implementation CRUDLayer (WOAppGoodies)
- (WOComponentController *)pageWithName:(NSString *)aName forEntity:(EOEntity*)anEntity
{
	WOComponentController *nextPage;
	
	nextPage = [modelData
			pageWithName:aName
			forEntityNamed:[anEntity name]];

	if (!nextPage)
		nextPage = [[anEntity classNameClass] pageWithName:aName];

	if (!nextPage)
		nextPage = [WOApp pageWithName:aName];

	return nextPage; 
}

- (WOComponentController *)pageWithName:(NSString *)aName forObject:(id)anObject
{
	WOComponentController *nextPage;
	
	nextPage = [modelData
			pageWithName:aName
			// TODO: This will cause subclasses of NSObject to break!
			forEntityNamed:[[anObject entity] name]];
	
	if (!nextPage)
		nextPage = [anObject pageWithName:aName];

	if (!nextPage)
		nextPage = [WOApp pageWithName:aName];

	return nextPage;
}
@end

@implementation CRUDLayer (EOAccessLayerAdditions)
- (NSNumber*)widthOfProperty:(NSString*)property onEntity:(EOEntity*)entity
{
	NSNumber *aValue;
	
	aValue = [modelData
			widthOfProperty:property
			onEntity:entity];

	if (!aValue)
		aValue = [[entity classNameClass]
				perform:@selector(widthOfProperty:)
				withObject:property];
	
	if (!aValue)
		aValue = [entity widthOfProperty:property];

	return aValue;
}

- (NSNumber*)heightOfProperty:(NSString*)property onEntity:(EOEntity*)entity
{
	NSNumber *aValue;
	
	aValue = [modelData
			heightOfProperty:property
			onEntity:entity];
	if (!aValue)
		aValue = [[entity classNameClass]
				perform:@selector(heightOfProperty:)
				withObject:property];
	
	if (!aValue)
		aValue = [entity heightOfProperty:property];

	return aValue;
}
@end

@implementation CRUDLayer (PropertyHandling)
- (NSArray*)_propertiesForEntity:(EOEntity*)entity forFunctionName:(NSString*)name andSelector:(SEL)selector
{
	NSArray *properties;
	
	properties = [modelData
			pageProperties:name
			forEntity:entity];

	if (!properties)
		properties = [[entity classNameClass] perform:selector];
	
	if (!properties)
		properties = [entity visibleAttributeNames];

	return properties;
}

- (NSArray*)_propertiesForObject:(id)anObject forFunctionName:(NSString*)name andSelector:(SEL)selector
{
	NSArray *properties;
	
	properties = [modelData
			pageProperties:name
			// TODO: This will cause subclasses of NSObject to break!!!
			forEntity:[anObject entity]];

	if (!properties)
		properties = [anObject perform:selector];
	
	if (!properties)
		// TODO: This will cause subclasses of NSObject to break!!!
		properties = [[anObject entity] visibleAttributeNames];

	return properties;
}


- (NSArray*)qualifierBuilderPropertiesForEntity:(EOEntity *)anEntity
{
    return [self
		_propertiesForEntity:anEntity
		   forFunctionName:@"qualifierBuilderProperties"
		       andSelector:@selector(qualifierBuilderProperties)];
}

- (NSArray*)sortablePropertiesForEntity:(EOEntity *)anEntity
{
    return [self
		_propertiesForEntity:anEntity
		   forFunctionName:@"sortableProperties"
		       andSelector:@selector(sortableProperties)];
}

	// Methods which work on an instance...
- (NSArray*)pickListPropertiesForEntity:(EOEntity *)anEntity
{
    return [self
		_propertiesForEntity:anEntity
		   forFunctionName:@"pickListProperties"
		       andSelector:@selector(pickListProperties)];
}

- (NSArray*)visiblePropertiesForObject:(id)anObject
{
    return [self
		_propertiesForObject:anObject
		   forFunctionName:@"visibleProperties"
		       andSelector:@selector(visibleProperties)];
}

- (NSArray*)editablePropertiesForObject:(id)anObject
{
    return [self
		_propertiesForObject:anObject
		   forFunctionName:@"editableProperties"
		       andSelector:@selector(editableProperties)];
}

- (BOOL)objectIsEditable:(id)anObject
{
    // TODO:  Need to call the object, etc
    // ...similar to the sequence of orders for other
    // api available through the CRUDLayer...
    
    // How to differentiate between NO and nil???
    
    return [modelData objectIsEditable:anObject];
}


	//
	// Another sick hack...(sigh)
	// I think the evolution process is bleeding at this point...
	//
- (NSArray*)sortColumnNamesForProperties:(NSArray*)properties
{
	NSMutableArray *columnNames = [NSMutableArray array];
	int count = [properties count];
	int index = 0;
	NSString *columnName;
	
	// uggg...
#define ALLTHATSTUFF \
	columnName = [NSString stringWithFormat:@"%d_%@", \
			index, 	[properties objectAtIndex:index]]; \
	[columnNames addObject:columnName]; \
	++index

	switch (count) {
	    default:
	//	ALLTHATSTUFF;
	    case 6:
	    case 5:
	    case 4:
		ALLTHATSTUFF;
	    case 3:
	    case 2:
		ALLTHATSTUFF;
	    case 1:
		ALLTHATSTUFF;
		break;
	    case 0:
	    	NSLog(@"sortColumnNamesForProperties: Hmmm...shouldn't be here...");
		break;
	}
	return columnNames;
}
@end

#ifdef CRUD_DEBUG
@implementation CRUDLayer (DebuggingDelegateMethods)
- (EODelegateResponse)adaptorChannel:channel willEvaluateExpression:(NSMutableString *)anExpression
{
	NSLog(@"DEBUG: INFO: adaptorChannel:%@: anExpression:'%@'", channel, anExpression);
	return EODelegateApproves;
}

- (BOOL)adaptor:(EOAdaptor *)adaptor willReportError:(NSString *)error
{
	NSLog(@"DEBUG: INFO: adaptor:'%@' willReportError:'%@'", [adaptor name], error);
	return NO;
}
@end
#endif
